#include <alsa/asoundlib.h>
#include <sys/select.h>

static snd_pcm_t *h_play;
static snd_pcm_t *h_record;
static int arec;
static int aplay;

static void fd_play(void)
{
#ifdef DO_NOTHING
  return;
#endif
  int count = snd_pcm_poll_descriptors_count(h_play);
  struct pollfd p;
  if (count != 1) R;
  if (snd_pcm_poll_descriptors(h_play, &p, count) < 0) R;
  aplay = p.fd;
}

static void fd_record(void)
{
#ifdef DO_NOTHING
  return;
#endif
  int count = snd_pcm_poll_descriptors_count(h_record);
  struct pollfd p;
  if (count != 1) R;
  if (snd_pcm_poll_descriptors(h_record, &p, count) < 0) R;
  arec = p.fd;
}

static void close_alsa(void)
{
#ifdef DO_NOTHING
  return;
#endif
  snd_pcm_close(h_play);
  snd_pcm_close(h_record);
}

static void open_alsa(void)
{
#ifdef DO_NOTHING
  return;
#endif
  int err;
  snd_pcm_hw_params_t *params;
  char *record_alsa_device = "hw:2";
  //char *play_alsa_device = "default:0";
  char *play_alsa_device = "hw:0";
  const char *devname_play = play_alsa_device;
  const char *devname_record = record_alsa_device;
  unsigned int rate;
  snd_pcm_uframes_t period_frames;
  snd_pcm_uframes_t buffer_frames;
  int record_format;

  h_play = 0;
  h_record = 0;

  /* play */

  snd_pcm_hw_params_alloca(&params);

  if ((err = snd_pcm_open(&h_play, devname_play,
                          SND_PCM_STREAM_PLAYBACK, 0)) < 0) R;
  if ((err = snd_pcm_hw_params_any(h_play, params)) < 0) R;
  if ((err = snd_pcm_hw_params_set_access(h_play, params,
                 SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) R;
  if ((err = snd_pcm_hw_params_set_format(h_play, params,
                 SND_PCM_FORMAT_S16_LE)) < 0) R;
  if ((err = snd_pcm_hw_params_set_channels(h_play, params, 2)) < 0) R;
  rate = 44100;
  if ((err = snd_pcm_hw_params_set_rate_near(h_play, params, &rate, 0)) < 0) R;
  if (rate!=44100) R;
  period_frames = SAMPLES;
  buffer_frames = SAMPLES*2;
  if ((err = snd_pcm_hw_params_set_period_size_near(h_play, params,
                 &period_frames, 0)) < 0) R;
  if ((err = snd_pcm_hw_params_set_buffer_size_near(h_play, params,
                 &buffer_frames)) < 0) R;
  if (period_frames != SAMPLES || buffer_frames != SAMPLES*2) {
    printf("got: period_frames %lu buffer_frames %lu\n", period_frames,
           buffer_frames);
    R;
  }
  if ((err = snd_pcm_hw_params(h_play, params)) < 0) R;

  /* record */

  if ((err = snd_pcm_open(&h_record, devname_record,
                          SND_PCM_STREAM_CAPTURE, 0)) < 0) R;
  if ((err = snd_pcm_hw_params_any(h_record, params)) < 0) R;
  if ((err = snd_pcm_hw_params_set_access(h_record, params,
                 SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) R;
  record_format = SND_PCM_FORMAT_S24_3LE;
  if ((err = snd_pcm_hw_params_set_format(h_record, params,
                 record_format)) < 0) R;
  if ((err = snd_pcm_hw_params_set_channels(h_record, params, 2)) < 0) R;
  rate = 44100;
  if ((err = snd_pcm_hw_params_set_rate_near(h_record, params, &rate, 0)) < 0)
    R;
  if (rate != 44100) R;
  period_frames = SAMPLES;
  buffer_frames = 65536;
  if ((err = snd_pcm_hw_params_set_period_size_near(h_record, params,
                 &period_frames, 0)) < 0) R;
  if ((err = snd_pcm_hw_params_set_buffer_size_near(h_record, params,
                 &buffer_frames)) < 0) R;
  if (period_frames != SAMPLES || buffer_frames != 65536) {
    printf("got: period_frames %lu buffer_frames %lu\n", period_frames,
           buffer_frames);
    R;
  }
  if ((err = snd_pcm_hw_params(h_record, params)) < 0) R;
  if ((err = snd_pcm_start(h_record)) < 0) R;
  fd_play();
  fd_record();
}

static void record_sound(void *b)
{
#ifdef DO_NOTHING
  return;
#endif
  size_t count = SAMPLES;
  ssize_t r;
  unsigned char *bb = b;

  while (count) {
    r = snd_pcm_readi(h_record, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) R;
    else if (r == -ESTRPIPE) R;
    else if (r < 0) R;
    if (r > 0) {
      count -= r;
      bb += r * 6; /* record_bytes_per_frame */
    }
  }
}

static void xrun(void)
{
  fprintf(stderr, "+"); fflush(stderr);
  snd_pcm_prepare(h_play);
}

static void suspend(void)
{
  int res;
  int loop = 0;
  while ((res = snd_pcm_resume(h_play)) == -EAGAIN && loop < 100) {
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10 * 1000;
    select(0, 0, 0, 0, &t);
    loop++;
  }
  if (res < 0) snd_pcm_prepare(h_play);
}

static void play_sound(void *b)
{
#ifdef DO_NOTHING
  return;
#endif
  size_t count = SAMPLES;
  ssize_t r;
//  int xrun_count = 0;
  unsigned char *bb = b;

#if 0
  while (count && xrun_count != 100) {
    r = snd_pcm_writei(h_play, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) { xrun(); xrun_count++; }
    else if (r == -ESTRPIPE) { suspend(); xrun_count++; }
    else if (r < 0) R;
    if (r > 0) {
      count -= r;
      bb += r * 4; /* play_bytes_per_frame */
    }
  }
  if (xrun_count) { play_sound(b); play_sound(b); }
#endif
  r = snd_pcm_writei(h_play, bb, count);
  if (r == -EPIPE) { xrun(); snd_pcm_writei(h_play, bb, count); }
  if (r == -ESTRPIPE) suspend();
}

static int more64(void)
{
#ifdef DO_NOTHING
  return 0;
#endif
  return snd_pcm_avail(h_record) >= SAMPLES;
}

#if 0
static int out64(void)
{
  return snd_pcm_avail(h_play);
}
#endif
